"""
This module contains the list of possible encodings, taken from
the standard library documentation:
https://docs.python.org/3.4/library/codecs.html#standard-encodings

"""

#: Encodings map, map a codec name to a an alias/language pair.#:
ENCODINGS_MAP = {
    "ascii": ("US-ASCII", "English"),
    "big5": ("Big5", "Chinese traditional"),
    "big5hkscs": ("Big5-HKSCS", "Chinese traditional"),
    "cp037": ("IBM037", "English"),
    "cp424": ("IBM242", "Hebrew"),
    "cp437": ("IBM437", "English"),
    "cp500": ("IBM500", "Western"),
    "cp737": ("IBM737", "Greek"),
    "cp775": ("IBM775", "Baltic"),
    "cp850": ("IBM850", "Western"),
    "cp852": ("IBM852", "Central European"),
    "cp855": ("IBM855", "Cyrillic"),
    "cp856": ("IBM856", "Hebrew"),
    "cp857": ("IBM857", "Turkish"),
    "cp860": ("IBM860", "Portugese"),
    "cp861": ("IBM861", "Icelandic"),
    "cp862": ("IBM862", "Hebrew"),
    "cp863": ("IBM863", "Canadian"),
    "cp864": ("IBM864", "Arabic"),
    "cp865": ("IBM865", "Nordic"),
    "cp866": ("IBM866", "Russian"),
    "cp869": ("IBM869", "Greek"),
    "cp874": ("IBM874", "Thai"),
    "cp875": ("IBM875", "Greek"),
    "cp932": ("IBM932", "Japanese"),
    "cp949": ("IBM949", "Korean"),
    "cp950": ("IBM950", "Chinese traditional"),
    "cp1006": ("IBM1006", "Urdu"),
    "cp1026": ("IBM1026", "Turkish"),
    "cp1140": ("IBM1140", "Western"),
    "cp1250": ("windows-1250", "Central European"),
    "cp1251": ("windows-1251", "Cyrillic"),
    "cp1252": ("windows-1252", "Western"),
    "cp1253": ("windows-1253", "Greek"),
    "cp1254": ("windows-1254", "      Turkish"),
    "cp1255": ("windows-1255", "Hebrew"),
    "cp1256": ("windows-1256", "      Arabic"),
    "cp1257": ("windows-1257", "Baltic"),
    "cp1258": ("windows-1258", "Vietnamese"),
    "euc_jp": ("EUC-JP", "Japanese"),
    "euc_jis_2004": ("EUC-JIS-2004", "Japanese"),
    "euc_jisx0213": ("EUC-JISX0213", "Japanese"),
    "euc_kr": ("EUC-KR", "Korean"),
    "gb2312": ("GB2312", "Chinese simplified"),
    "gbk": ("GBK", "Chinese unified"),
    "gb18030": ("GB18030", "Chinese unified"),
    "hz": ("HZ", "Chinese simplified"),
    "iso2022_jp": ("ISO-2022-JP", "Japanese"),
    "iso2022_jp_1": ("ISO-2022-JP-1", "Japanese"),
    "iso2022_jp_2": ("ISO-2022-JP-2", "Japanese"),
    "iso2022_jp_2004": ("ISO-2022-JP-2004", "Japanese"),
    "iso2022_jp_3": ("ISO-2022-JP-3", "Japanese"),
    "iso2022_jp_ext": ("ISO-2022-JP-EXT", "Japanese"),
    "iso2022_kr": ("ISO-2022-KR", "Korean"),
    "latin_1": ("ISO-8859-1", "Western"),
    "iso8859_2": ("ISO-8859-2", "Central European"),
    "iso8859_3": ("ISO-8859-3", "South European"),
    "iso8859_4": ("ISO-8859-4", "Baltic"),
    "iso8859_5": ("ISO-8859-5", "Cyrillic"),
    "iso8859_6": ("ISO-8859-6", "Arabic"),
    "iso8859_7": ("ISO-8859-7", "Greek"),
    "iso8859_8": ("ISO-8859-8", "Hebrew"),
    "iso8859_9": ("ISO-8859-9", "Turkish"),
    "iso8859_10": ("ISO-8859-10", "Nordic"),
    "iso8859_13": ("ISO-8859-13", "Baltic"),
    "iso8859_14": ("ISO-8859-14", "Celtic"),
    "iso8859_15": ("ISO-8859-15", "Western"),
    "johab": ("Johab", "Korean"),
    "koi8_r": ("KOI8-R", "Russian"),
    "koi8_u": ("KOI8-U", "Ukrainian"),
    "mac_cyrillic": ("MacCyrillic", "Cyrillic"),
    "mac_greek": ("MacGreek", "Greek"),
    "mac_iceland": ("MacIceland", "Icelandic"),
    "mac_latin2": ("MacCentralEurope", "Central European"),
    "mac_roman": ("MacRoman", "Western"),
    "mac_turkish": ("MacTurkish", "Turkish"),
    "ptcp154": ("PTCP154", "Cyrillic Asian"),
    "shift_jis": ("Shift_JIS", "Japanese"),
    "shift_jis_2004": ("Shift_JIS-2004", "Japanese"),
    "shift_jisx0213": ("Shift_JISX0213", "Japanese"),
    "utf_16": ("UTF-16", "Unicode"),
    "utf_16_be": ("UTF-16BE", "Unicode"),
    "utf_16_le": ("UTF-16LE", "Unicode"),
    "utf_7": ("UTF-7", "Unicode"),
    "utf_8": ("UTF-8", "Unicode")
}


def convert_to_codec_key(value):
    """
    Normalize code key value (encoding codecs must be lower case and must
    not contain any dashes).

    :param value: value to convert.
    """
    if not value:
        # fallback to utf-8
        value = 'UTF-8'
    # UTF-8 -> utf_8
    converted = value.replace('-', '_').lower()
    # fix some corner cases, see https://github.com/pyQode/pyQode/issues/11
    all_aliases = {
        'ascii': [
            'us_ascii',
            'us',
            'ansi_x3.4_1968',
            'cp367',
            'csascii',
            'ibm367',
            'iso_ir_6',
            'iso646_us',
            'iso_646.irv:1991'
        ],
        'utf-7': [
            'csunicode11utf7',
            'unicode_1_1_utf_7',
            'unicode_2_0_utf_7',
            'x_unicode_1_1_utf_7',
            'x_unicode_2_0_utf_7',
        ],
        'utf_8': [
            'unicode_1_1_utf_8',
            'unicode_2_0_utf_8',
            'x_unicode_1_1_utf_8',
            'x_unicode_2_0_utf_8',
        ],
        'utf_16': [
            'utf_16le',
            'ucs_2',
            'unicode',
            'iso_10646_ucs2'
        ],
        'latin_1': ['iso_8859_1']
    }

    for key, aliases in all_aliases.items():
        if converted in aliases:
            return key
    return converted



import glob
import re
import pytest
import sys
from ..helpers import delete_file_on_return, editor_open
from pyqode.core.api import folding, TextBlockHelper, TextHelper
from pyqode.qt.QtTest import QTest


class FoldDetectorTestCase(object):
    """
    Checks that the fold detector detects the corret layout when loading
    files (or setting text). Dynamic checks (i.e. when the user edit the text
    are performed by DynamicFoldDetectorTestCase).
    """
    def __init__(self, test_file, results_file):
        with open(test_file, 'r') as f:
            self.test_file_content = f.read()
        with open(results_file, 'r') as f:
            self.expected_results_content = f.read()

    @delete_file_on_return('file_structure')
    def execute(self, editor):
        editor.setPlainText(self.test_file_content, '', '')
        with open('file_structure', 'w') as f:
            folding.print_tree(editor, file=f, print_blocks=True)
        with open('file_structure', 'r') as f:
            results_content = f.read()
        assert results_content == self.expected_results_content


@pytest.mark.parametrize('case', [
    FoldDetectorTestCase('test/test_api/folding_cases/foo.py',
                         'test/test_api/folding_cases/foo.static_results')
])
def test_fold_detection_static(editor, case):
    case.execute(editor)


def test_fold_limit(editor):
    editor.syntax_highlighter.fold_detector.limit = 1
    editor.file.open('test/test_api/folding_cases/foo.py')
    block = editor.document().firstBlock()
    while block.blockNumber() < editor.blockCount() - 1:
        assert TextBlockHelper.get_fold_lvl(block) <= 1
        block = block.next()
    editor.syntax_highlighter.fold_detector.limit = sys.maxsize


def test_base_fold_indenter():
    f = folding.FoldDetector()
    with pytest.raises(NotImplementedError):
        f.detect_fold_level(None, None)


class DynamicFoldDetectorTestCase(object):
    """
    A dynamic test consists in loading a file and performing a few predefined
    actions (such as going to a specific line/column, pressing a key) and then
    comparing the results with the expected results (this part is similar to
    the static test cases)

    The case will load a context file which must contains the following three
    entries:
        - input: path relative to the root of the source dir
        - results: path relative to the root of the source dir
        - actions: one or more actions to perform (see the list below).
             each action is separated by a ';' and must be on the same line

    Here are the actions that you can perform:

        - goto_line(x, y): goto line x and column y
        - press(key): simulate a key press at the current cursor location.
                      key is the key value (e.g int(QtCore.Qt.Key_A) )
        - fold: fold the current block
        - unfold: unfold the current block
        - assert_parent: assert scope has a parent
        - assert_no_parent: assert scope has not parent
        - assert_nb_children(x): assert nb children equals x
        - assert_not_scope_start: assert the block is not the start of a fold scope

    Example::

        input: test/test_api/folding_cases/foo.py
        results: test/test_api/folding_cases/foo.dynamic_results_2
        actions:goto_line(9,0);press(16777220);fold();unfold()


    """
    def __init__(self, ctx_file):
        self._name = ctx_file
        self.file_content = None
        self.expected_results_file_content = None
        self.actions = []
        with open(ctx_file, 'r') as ctx_f:
            for l in ctx_f.read().splitlines():
                if len(l.split(':')) == 2:
                    section, value = l.split(':')
                    if section == 'input':
                        with open(value.strip(), 'r') as f:
                            self.file_content = f.read()
                    elif section == 'results':
                        with open(value.strip(), 'r') as f:
                            self.expected_results_file_content = f.read()
                    elif section == 'actions':
                        self.actions = value.strip().split(';')
        assert (len(self.file_content) and
                len(self.expected_results_file_content) and
                len(self.actions))

    def _get_scope(self, editor):
        try:
            scope = folding.FoldScope(editor.textCursor().block())
        except ValueError:
            return None
        else:
            assert len(scope.text())
            return scope

    def perform_actions(self, editor):
        for action in self.actions:
            if action.startswith('goto_line('):
                p = re.compile(r'\b\d+\b')
                matches = list(p.finditer(action))
                assert len(matches) == 2
                line = int(matches[0].group())
                column = int(matches[1].group())
                TextHelper(editor).goto_line(line - 1, column)
            elif action.startswith('press('):
                p = re.compile(r'\b\d+\b')
                matches = list(p.finditer(action))
                assert len(matches) == 1
                key = int(matches[0].group())
                QTest.keyPress(editor, key)
            elif action.startswith('assert_nb_children('):
                p = re.compile(r'\b\d+\b')
                matches = list(p.finditer(action))
                assert len(matches) == 1
                nb = int(matches[0].group())
                scope = self._get_scope(editor)
                assert len(list(scope.child_regions())) == nb
            elif action == 'fold':
                scope = self._get_scope(editor)
                scope.fold()
                assert scope.collapsed
            elif action == 'unfold':
                scope = self._get_scope(editor)
                scope.unfold()
                assert not scope.collapsed
            elif action == 'assert_parent':
                scope = self._get_scope(editor)
                assert scope.parent() is not None
            elif action == 'assert_no_parent':
                scope = self._get_scope(editor)
                assert scope.parent() is None
            elif action == 'assert_not_scope_start':
                scope = self._get_scope(editor)
                assert scope is None

    @delete_file_on_return('file_structure')
    def execute(self, editor):
        editor.setPlainText(self.file_content, '', '')
        self.perform_actions(editor)
        with open('file_structure', 'w') as f:
            folding.print_tree(editor, file=f, print_blocks=True)
        # Compare results
        with open('file_structure', 'r') as f:
            results_content = f.read()
        assert results_content == self.expected_results_file_content

    def __repr__(self):
        return self._name


@pytest.mark.parametrize('case', [
    DynamicFoldDetectorTestCase(f) for f
    in sorted(glob.glob('test/test_api/folding_cases/*.ctx'))
])
def test_fold_detection_dynamic(editor, case):
    case.execute(editor)



"""
This example will show you that the caret line color is automatically set
by the syntax highlighter mode.

See how the color is different from the color obtained in
``caret_line_highligter.py``
"""
import logging
logging.basicConfig(level=logging.DEBUG)
import sys

from pyqode.qt import QtWidgets
from pyqode.core.api import CodeEdit
from pyqode.core.backend import server
from pyqode.core.modes import CaretLineHighlighterMode
from pyqode.core.modes import PygmentsSyntaxHighlighter


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    editor = CodeEdit()
    editor.backend.start(server.__file__)
    editor.resize(800, 600)
    print(editor.modes.append(CaretLineHighlighterMode()))
    print(editor.modes.append(PygmentsSyntaxHighlighter(editor.document())))
    editor.show()
    editor.appendPlainText('\n' * 10)
    app.exec_()
    editor.close()
    del editor
    del app



"""
Show how decoding errors can be automatically handled using the EncodingPanel.
"""
import sys
from pyqode.qt import QtWidgets
from common import setup_editor, get_file_path

app = QtWidgets.QApplication(sys.argv)
editor = setup_editor()
editor.file.open(get_file_path(), use_cached_encoding=False)
app.exec_()



"""
This module contains the panel API.
"""
import logging
from pyqode.core.api.mode import Mode
from pyqode.qt import QtWidgets, QtGui


def _logger():
    """ Returns module's logger """
    return logging.getLogger(__name__)


class Panel(QtWidgets.QWidget, Mode):
    """
    Base class for editor panels.

    A panel is a mode and a QWidget.

    .. note:: Use enabled to disable panel actions and setVisible to change the
        visibility of the panel.
    """

    _use_syntax_theme = False

    class Position(object):
        """
        Enumerates the possible panel positions
        """
        #: Top margin
        TOP = 0
        #: Left margin
        LEFT = 1
        #: Right margin
        RIGHT = 2
        #: Bottom margin
        BOTTOM = 3

        @classmethod
        def iterable(cls):
            """ Returns possible positions as an iterable (list) """
            return [cls.TOP, cls.LEFT, cls.RIGHT, cls.BOTTOM]

    @property
    def scrollable(self):
        """
        A scrollable panel will follow the editor's scroll-bars. Left and right
        panels follow the vertical scrollbar. Top and bottom panels follow the
        horizontal scrollbar.

        :type: bool
        """
        return self._scrollable

    @scrollable.setter
    def scrollable(self, value):
        self._scrollable = value

    def __init__(self, dynamic=False):
        Mode.__init__(self)
        QtWidgets.QWidget.__init__(self)
        #: Specifies whether the panel is dynamic. A dynamic panel is a panel
        #: that will be shown/hidden depending on the context.
        #: Dynamic panel should not appear in any GUI menu (e.g. no display
        #: in the panels menu of the notepad example).
        self.dynamic = dynamic
        #: Panel order into the zone it is installed to. This value is
        #: automatically set when installing the panel but it can be changed
        #: later (negative values can also be used).
        self.order_in_zone = -1
        self._scrollable = False
        #: Position in the editor (top, left, right, bottom)
        self.position = -1
        self._cached_background_brush = None
        self._cached_foreground_pen = None

    def on_install(self, editor):
        """
        Extends :meth:`pyqode.core.api.Mode.on_install` method to set the
        editor instance as the parent widget.

        .. warning:: Don't forget to call **super** if you override this
            method!

        :param editor: editor instance
        :type editor: pyqode.core.api.CodeEdit
        """
        Mode.on_install(self, editor)
        self.setParent(editor)
        self.setPalette(QtWidgets.QApplication.instance().palette())
        self.setFont(QtWidgets.QApplication.instance().font())
        self.editor.panels.refresh()

    @property
    def _background_brush(self):

        """
        Gives the brush that is used to draw the background. The panel can
        adopt the syntax theme, so that it blends in with the editor, which
        looks best for example for line numbers, or it can adopt the
        application theme, which looks best for example for the search panel.
        The brush is cached for performance.
        """

        if self._cached_background_brush is not None:
            return self._cached_background_brush
        if not self._use_syntax_theme:
            self._cached_background_brush = QtGui.QBrush(
                QtGui.QColor(self.palette().window().color())
            )
        else:
            try:
                self._color_scheme = \
                    self.editor.syntax_highlighter.color_scheme
            except AttributeError:
                # There is no syntax highlighter to adopt the theme from
                self._use_syntax_theme = False
                return self._background_brush
        self._cached_background_brush = \
            self._color_scheme.formats['background'].background()
        return self._cached_background_brush

    @property
    def _foreground_pen(self):

        """
        Gives the pen that is used to draw the foreground. The panel can
        adopt the syntax theme, so that it blends in with the editor, which
        looks best for example for line numbers, or it can adopt the
        application theme, which looks best for example for the search panel.
        The pen is cached for performance.
        """

        if self._cached_foreground_pen is not None:
            return self._cached_foreground_pen
        if not self._use_syntax_theme:
            self._cached_foreground_pen = QtGui.QPen(
                QtGui.QColor(self.palette().windowText().color())
            )
        else:
            try:
                self._color_scheme = \
                    self.editor.syntax_highlighter.color_scheme
            except AttributeError:
                # There is no syntax highlighter to adopt the theme from
                self._use_syntax_theme = False
                return self._foreground_pen
        self._cached_foreground_pen = \
            self._color_scheme.formats['normal'].foreground()
        return self._cached_foreground_pen

    def paintEvent(self, event):

        if not self.isVisible():
            return
        QtGui.QPainter(self).fillRect(event.rect(), self._background_brush)

    def setVisible(self, visible):
        """
        Shows/Hides the panel

        Automatically call CodeEdit.refresh_panels.

        :param visible: Visible state
        """
        _logger().log(5, '%s visibility changed', self.name)
        super(Panel, self).setVisible(visible)
        if self.editor:
            self.editor.panels.refresh()



"""
Contains a panel to manage unicode decoding/encoding errors.

"""
from pyqode.core.api.panel import Panel
from pyqode.core.api.decoration import TextDecoration
from pyqode.qt import QtCore, QtGui, QtWidgets


class ReadOnlyPanel(Panel):
    """ Displays a message if the opened file is read-only """

    @property
    def color(self):
        """
        Returns the panel color.
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._refresh_stylesheet()
        if self.editor:
            # propagate changes to every clone
            for clone in self.editor.clones:
                try:
                    clone.modes.get(self.__class__).color = value
                except KeyError:
                    # this should never happen since we're working with clones
                    pass

    @property
    def foreground(self):
        return self._foreground

    @foreground.setter
    def foreground(self, value):
        self._foreground = value
        self._refresh_stylesheet()
        # propagate changes to every clone
        if self.editor:
            for clone in self.editor.clones:
                try:
                    clone.modes.get(self.__class__).foreground = value
                except KeyError:
                    # this should never happen since we're working with clones
                    pass

    def _refresh_stylesheet(self):
        try:
            self._lbl_stylesheet = ('color: %s;background: %s' %
                                    (self._foreground.name(),
                                     self._color.name()))
            self.ui.lblDescription.setStyleSheet(self._lbl_stylesheet)
        except AttributeError:
            pass

    def __init__(self):
        super(ReadOnlyPanel, self).__init__(dynamic=True)
        # leave it here otherwise you will have circular import errors
        from pyqode.core._forms.pnl_read_only_ui import Ui_Form
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self._color = None
        self.color = QtGui.QColor('#8AADD4')
        self._foreground = None
        self.foreground = QtGui.QColor('#FFFFFF')
        self.hide()

    @property
    def _background_brush(self):
        return QtGui.QBrush(self._color)
